/* specfunc/gdl_sf_exp.h
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2004 Gerard Jungman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Author:  G. Jungman */

#ifndef __GDL_SF_EXP_H__
#define __GDL_SF_EXP_H__

#include <gdl/gdl_common.h>
#include <gdl/gdl_sf_result.h>
#include <gdl/gdl_precision.h>

__BEGIN_DECLS

/* Provide an exp() function with GDL semantics,
 * i.e. with proper error checking, etc.
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_e(const double x, gdl_sf_result * result);
double gdl_sf_exp(const double x);


/* Exp(x)
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_e10_e(const double x, gdl_sf_result_e10 * result);


/* Exponentiate and multiply by a given factor:  y * Exp(x)
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_mult_e(const double x, const double y, gdl_sf_result * result);
double gdl_sf_exp_mult(const double x, const double y);


/* Exponentiate and multiply by a given factor:  y * Exp(x)
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_mult_e10_e(const double x, const double y, gdl_sf_result_e10 * result);


/* exp(x)-1
 *
 * exceptions: GDL_EOVRFLW
 */
int gdl_sf_expm1_e(const double x, gdl_sf_result * result);
double gdl_sf_expm1(const double x);


/* (exp(x)-1)/x = 1 + x/2 + x^2/(2*3) + x^3/(2*3*4) + ...
 *
 * exceptions: GDL_EOVRFLW
 */
int gdl_sf_exprel_e(const double x, gdl_sf_result * result);
double gdl_sf_exprel(const double x);


/* 2(exp(x)-1-x)/x^2 = 1 + x/3 + x^2/(3*4) + x^3/(3*4*5) + ...
 *
 * exceptions: GDL_EOVRFLW
 */
int gdl_sf_exprel_2_e(double x, gdl_sf_result * result);
double gdl_sf_exprel_2(const double x);


/* Similarly for the N-th generalization of
 * the above. The so-called N-relative exponential
 *
 * exprel_N(x) = N!/x^N (exp(x) - Sum[x^k/k!, {k,0,N-1}])
 *             = 1 + x/(N+1) + x^2/((N+1)(N+2)) + ...
 *             = 1F1(1,1+N,x)
 */
int gdl_sf_exprel_n_e(const int n, const double x, gdl_sf_result * result);
double gdl_sf_exprel_n(const int n, const double x);


/* Exponentiate a quantity with an associated error.
 */
int gdl_sf_exp_err_e(const double x, const double dx, gdl_sf_result * result);

/* Exponentiate a quantity with an associated error.
 */
int gdl_sf_exp_err_e10_e(const double x, const double dx, gdl_sf_result_e10 * result);


/* Exponentiate and multiply by a given factor:  y * Exp(x),
 * for quantities with associated errors.
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_mult_err_e(const double x, const double dx, const double y, const double dy, gdl_sf_result * result);


/* Exponentiate and multiply by a given factor:  y * Exp(x),
 * for quantities with associated errors.
 *
 * exceptions: GDL_EOVRFLW, GDL_EUNDRFLW
 */
int gdl_sf_exp_mult_err_e10_e(const double x, const double dx, const double y, const double dy, gdl_sf_result_e10 * result);

__END_DECLS


#ifdef HAVE_INLINE
#include <gdl/gdl_math.h>
#include <gdl/gdl_errno.h>

__BEGIN_DECLS

extern inline
int gdl_sf_exp_e(const double x, gdl_sf_result * result)
{
  if(x > GDL_LOG_DBL_MAX) {
    result->val = GDL_POSINF;
    result->err = GDL_POSINF;
    GDL_ERROR ("overflow", GDL_EOVRFLW);
  }
  else if(x < GDL_LOG_DBL_MIN) {
    result->val = 0.0;
    result->err = GDL_DBL_MIN;
    GDL_ERROR ("underflow", GDL_EUNDRFLW);
  }
  else {
    result->val = exp(x);
    result->err = 2.0 * GDL_DBL_EPSILON * result->val;
    return GDL_SUCCESS;
  }  
}

__END_DECLS

#endif /* HAVE_INLINE */

#endif /* __GDL_SF_EXP_H__ */
