/* vector/gdl_vector_char.h
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000 Gerard Jungman, Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GDL_VECTOR_CHAR_H__
#define __GDL_VECTOR_CHAR_H__

#include <stdlib.h>

#include <gdl/gdl_common.h>
#include <gdl/gdl_types.h>
#include <gdl/gdl_errno.h>
#include <gdl/gdl_check_range.h>
#include <gdl/gdl_gblock.h>

__BEGIN_DECLS

typedef struct 
{
  size_t size;
  size_t stride;
  char *data;
  gdl_block_char *block;
  int owner;
} 
gdl_vector_char;

typedef struct
{
  gdl_vector_char vector;
} _gdl_vector_char_view;

typedef _gdl_vector_char_view gdl_vector_char_view;

typedef struct
{
  gdl_vector_char vector;
} _gdl_vector_char_const_view;

typedef const _gdl_vector_char_const_view gdl_vector_char_const_view;


/* Allocation */

gdl_vector_char *gdl_vector_char_alloc (const size_t n);
gdl_vector_char *gdl_vector_char_calloc (const size_t n);

gdl_vector_char *gdl_vector_char_alloc_from_block (gdl_block_char * b,
                                                     const size_t offset, 
                                                     const size_t n, 
                                                     const size_t stride);

gdl_vector_char *gdl_vector_char_alloc_from_vector (gdl_vector_char * v,
                                                      const size_t offset, 
                                                      const size_t n, 
                                                      const size_t stride);

void gdl_vector_char_free (gdl_vector_char * v);

/* Views */

_gdl_vector_char_view 
gdl_vector_char_view_array (char *v, size_t n);

_gdl_vector_char_view 
gdl_vector_char_view_array_with_stride (char *base,
                                         size_t stride,
                                         size_t n);

_gdl_vector_char_const_view 
gdl_vector_char_const_view_array (const char *v, size_t n);

_gdl_vector_char_const_view 
gdl_vector_char_const_view_array_with_stride (const char *base,
                                               size_t stride,
                                               size_t n);

_gdl_vector_char_view 
gdl_vector_char_subvector (gdl_vector_char *v, 
                            size_t i, 
                            size_t n);

_gdl_vector_char_view 
gdl_vector_char_subvector_with_stride (gdl_vector_char *v, 
                                        size_t i,
                                        size_t stride,
                                        size_t n);

_gdl_vector_char_const_view 
gdl_vector_char_const_subvector (const gdl_vector_char *v, 
                                  size_t i, 
                                  size_t n);

_gdl_vector_char_const_view 
gdl_vector_char_const_subvector_with_stride (const gdl_vector_char *v, 
                                              size_t i, 
                                              size_t stride,
                                              size_t n);

/* Operations */

char gdl_vector_char_get (const gdl_vector_char * v, const size_t i);
void gdl_vector_char_set (gdl_vector_char * v, const size_t i, char x);

char *gdl_vector_char_ptr (gdl_vector_char * v, const size_t i);
const char *gdl_vector_char_const_ptr (const gdl_vector_char * v, const size_t i);

void gdl_vector_char_set_zero (gdl_vector_char * v);
void gdl_vector_char_set_all (gdl_vector_char * v, char x);
int gdl_vector_char_set_basis (gdl_vector_char * v, size_t i);

gdl_vector_char * gdl_vector_char_fread (FILE * stream);
int gdl_vector_char_fwrite (FILE * stream, const gdl_vector_char * v);
gdl_vector_char * gdl_vector_char_fscanf (FILE * stream);
int gdl_vector_char_fprintf (FILE * stream, const gdl_vector_char * v,
                              const char *format);

int gdl_vector_char_memcpy (gdl_vector_char * dest, const gdl_vector_char * src);

int gdl_vector_char_reverse (gdl_vector_char * v);

int gdl_vector_char_swap (gdl_vector_char * v, gdl_vector_char * w);
int gdl_vector_char_swap_elements (gdl_vector_char * v, const size_t i, const size_t j);

char gdl_vector_char_max (const gdl_vector_char * v);
char gdl_vector_char_min (const gdl_vector_char * v);
void gdl_vector_char_minmax (const gdl_vector_char * v, char * min_out, char * max_out);

size_t gdl_vector_char_max_index (const gdl_vector_char * v);
size_t gdl_vector_char_min_index (const gdl_vector_char * v);
void gdl_vector_char_minmax_index (const gdl_vector_char * v, size_t * imin, size_t * imax);

int gdl_vector_char_add (gdl_vector_char * a, const gdl_vector_char * b);
int gdl_vector_char_sub (gdl_vector_char * a, const gdl_vector_char * b);
int gdl_vector_char_mul (gdl_vector_char * a, const gdl_vector_char * b);
int gdl_vector_char_div (gdl_vector_char * a, const gdl_vector_char * b);
int gdl_vector_char_scale (gdl_vector_char * a, const double x);
int gdl_vector_char_add_constant (gdl_vector_char * a, const double x);

int gdl_vector_char_isnull (const gdl_vector_char * v);

#ifdef HAVE_INLINE

extern inline
char
gdl_vector_char_get (const gdl_vector_char * v, const size_t i)
{
#if GDL_RANGE_CHECK
  if (i >= v->size)
    {
      GDL_ERROR_VAL ("index out of range", GDL_EINVAL, 0);
    }
#endif
  return v->data[i * v->stride];
}

extern inline
void
gdl_vector_char_set (gdl_vector_char * v, const size_t i, char x)
{
#if GDL_RANGE_CHECK
  if (i >= v->size)
    {
      GDL_ERROR_VOID ("index out of range", GDL_EINVAL);
    }
#endif
  v->data[i * v->stride] = x;
}

extern inline
char *
gdl_vector_char_ptr (gdl_vector_char * v, const size_t i)
{
#if GDL_RANGE_CHECK
  if (i >= v->size)
    {
      GDL_ERROR_NULL ("index out of range", GDL_EINVAL);
    }
#endif
  return (char *) (v->data + i * v->stride);
}

extern inline
const char *
gdl_vector_char_const_ptr (const gdl_vector_char * v, const size_t i)
{
#if GDL_RANGE_CHECK
  if (i >= v->size)
    {
      GDL_ERROR_NULL ("index out of range", GDL_EINVAL);
    }
#endif
  return (const char *) (v->data + i * v->stride);
}


#endif /* HAVE_INLINE */

__END_DECLS

#endif /* __GDL_VECTOR_CHAR_H__ */


